#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <getopt.h>
#include <netdb.h>
#include <unistd.h>
#include <pthread.h>
#include <sys/socket.h>
#include <netinet/in.h>

unsigned short port = 2007;

struct in_addr *destinations = NULL;
unsigned num_destinations = 0;
unsigned room_destinations = 0;

struct in_addr *sources = NULL;
unsigned num_sources = 0;
unsigned room_sources = 0;

const static struct option longopts[] = {
	{ "source-list", required_argument, NULL, 's' },
	{ "destination-list", required_argument, NULL, 'd' },
	{ "port", required_argument, NULL, 'p' },
	{ NULL, 0, NULL, 0 }
};

void read_ip_list(char *filename, struct in_addr **addr_list, unsigned *num, unsigned *room)
{
	char buf[256];
	FILE *in = fopen(filename, "r");
	if (in == NULL) {
		perror(filename);
		exit(1);
	}

	for ( ;; ) {
		char *ptr;
		struct in_addr addr;
		struct hostent *he;

		if (fgets(buf, 256, in) == NULL)
			break;

		ptr = strchr(buf, '\n');
		if (ptr != NULL)
			*ptr = 0;

		ptr = strchr(buf, '\r');
		if (ptr != NULL)
			*ptr = 0;
		
		ptr = buf + strspn(buf, " \t");

		if (ptr[0] == '#' || ptr[0] == 0)
			continue;

		he = gethostbyname(ptr);
		if (he == NULL) {
			perror(ptr);
			exit(1);
		}

		// just pick the first for now
		memcpy(&addr.s_addr, he->h_addr_list[0], sizeof(addr.s_addr));

		if (*num >= *room) {
			if (*room == 0) {
				*room = 16;
			} else {
				*room <<= 1;
			}
			*addr_list = (struct in_addr *)realloc(*addr_list, *room * sizeof(struct in_addr));
		}

		(*addr_list)[*num] = addr;
		++*num;
	}

	fclose(in);
}

void parse_options(int argc, char **argv)
{
	int option_index = 0;

	for ( ;; ) {
		int c = getopt_long(argc, argv, "s:d:p:", longopts, &option_index); 
		switch (c) {
		case 's':
			read_ip_list(optarg, &sources, &num_sources, &room_sources);
			break;
		case 'd':
			read_ip_list(optarg, &destinations, &num_destinations, &room_destinations);
			break;
		case 'p':
			port = atoi(optarg);
			break;
		case -1:
			return;       // end of argument list
		default:
			fprintf(stderr, "Invalid option\n");
			exit(1);
		}
	}
}

void *receiver_worker(void *arg)
{
	int sock = (int)arg;
	char buf[65536];

	printf("Received worker for socket %u\n", sock);

	for ( ;; ) {
		int ret = read(sock, buf, 65536);
		if (ret == 0)
			break;

		// FIXME: update stats here
	}

	printf("Socket %u done\n", sock);
	
	if (close(sock) == -1) {
		perror("close()");
		exit(1);
	}

	pthread_exit(0);
}

int get_server_socket(unsigned short port)
{
	int server_sock;
	struct sockaddr_in sin;
	unsigned one = 1;

	server_sock = socket(PF_INET, SOCK_STREAM, IPPROTO_TCP);
	if (server_sock == -1) {
		perror("socket()");
		exit(1);
	}

	if (setsockopt(server_sock, SOL_SOCKET, SO_REUSEADDR, &one, sizeof(one)) == -1) {
		perror("setsocket(SO_REUSEADDR)");
		exit(1);
	}

	sin.sin_family = AF_INET;
	sin.sin_port = htons(port);
	sin.sin_addr.s_addr = INADDR_ANY;

	if (bind(server_sock, (struct sockaddr *)&sin, sizeof(struct sockaddr)) == -1) {
		perror("bind()");
		exit(1);
	}

	if (listen(server_sock, 255) == -1) {
		perror("listen()");
		exit(1);
	}

	return server_sock;
}

int main(int argc, char **argv)
{
	int server_sock;

	parse_options(argc, argv);

	if (destinations == NULL || sources == NULL) {
		fprintf(stderr, "Missing or empty source or destination host list. Aborting.\n");
	}

	server_sock = get_server_socket(port);
	
	printf("Sending data on port %u from %u sources to %u destinations.\n\n",
		port, num_sources, num_destinations);

	// FIXME: fire off sender workers here

	/*
	 * Listen for incoming connections, spawning off one receiver
	 * thread for each (which will just gobble up the data until
	 * we're done).
	 */
	for ( ;; ) {
		struct sockaddr_in addr;
		socklen_t addr_len = sizeof(addr);
		pthread_t thread;
		pthread_attr_t attr;

		int sock = accept(server_sock, (struct sockaddr *)&addr, &addr_len);
		if (sock == -1) {
			perror("accept()");
			exit(1);
		}

		// FIXME: these do not really set errno
		if (pthread_attr_init(&attr) != 0) {
			perror("pthread_attr_init()");
			exit(1);
		}

		if (pthread_attr_setstacksize(&attr, PTHREAD_STACK_MIN + 65536 + 0x4000) != 0) {
			perror("pthread_attr_setstacksize");
			exit(1);
		}

		if (pthread_create(&thread, &attr, receiver_worker, (void *)sock) != 0) {
			perror("pthread_create()");
			exit(1);
		}
	}

	exit(0);
}
